..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _cinder-quobyte:

.. list-table:: Description of Quobyte USP volume driver configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[DEFAULT]**
     -
   * - ``quobyte_client_cfg`` = ``None``
     - (String) Path to a Quobyte Client configuration file.
   * - ``quobyte_mount_point_base`` = ``$state_path/mnt``
     - (String) Base dir containing the mount point for the Quobyte volume.
   * - ``quobyte_qcow2_volumes`` = ``True``
     - (Boolean) Create volumes as QCOW2 files rather than raw files.
   * - ``quobyte_sparsed_volumes`` = ``True``
     - (Boolean) Create volumes as sparse files which take no space. If set to False, volume is created as regular file.In such case volume creation takes a lot of time.
   * - ``quobyte_volume_url`` = ``None``
     - (URI) URL to the Quobyte volume e.g., quobyte://<DIR host>/<volume name>
