..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _cinder-hgst:

.. list-table:: Description of HGST volume driver configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[DEFAULT]**
     -
   * - ``hgst_net`` = ``Net 1 (IPv4)``
     - (String) Space network name to use for data transfer
   * - ``hgst_redundancy`` = ``0``
     - (String) Should spaces be redundantly stored (1/0)
   * - ``hgst_space_group`` = ``disk``
     - (String) Group to own created spaces
   * - ``hgst_space_mode`` = ``0600``
     - (String) UNIX mode for created spaces
   * - ``hgst_space_user`` = ``root``
     - (String) User to own created spaces
   * - ``hgst_storage_servers`` = ``os:gbd0``
     - (String) Comma separated list of Space storage servers:devices. ex: os1_stor:gbd0,os2_stor:gbd0
