..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _cinder-zoning_fabric_brcd:

.. list-table:: Description of brocade zoning fabrics configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[BRCD_FABRIC_EXAMPLE]**
     -
   * - ``fc_fabric_address`` =
     - (String) Management IP of fabric.
   * - ``fc_fabric_password`` =
     - (String) Password for user.
   * - ``fc_fabric_port`` = ``22``
     - (Port number) Connecting port
   * - ``fc_fabric_ssh_cert_path`` =
     - (String) Local SSH certificate Path.
   * - ``fc_fabric_user`` =
     - (String) Fabric user ID.
   * - ``fc_southbound_protocol`` = ``HTTP``
     - (String) South bound connector for the fabric.
   * - ``fc_virtual_fabric_id`` = ``None``
     - (String) Virtual Fabric ID.
   * - ``principal_switch_wwn`` = ``None``
     - (String) DEPRECATED: Principal switch WWN of the fabric. This option is not used anymore.
   * - ``zone_activate`` = ``True``
     - (Boolean) Overridden zoning activation state.
   * - ``zone_name_prefix`` = ``openstack``
     - (String) Overridden zone name prefix.
   * - ``zoning_policy`` = ``initiator-target``
     - (String) Overridden zoning policy.
