.. -*- rst -*-

API versions
============




Show API v2 details
~~~~~~~~~~~~~~~~~~~

.. rest_method::  GET /v2/

Shows details for Block Storage API v2.


Normal response codes: 200


Request
-------



Response Parameters
-------------------

.. rest_parameters:: parameters.yaml

   - location: location



Response Example
----------------

.. literalinclude:: ./samples/version-show-response.json
   :language: javascript


