..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _cinder-lvm:

.. list-table:: Description of LVM configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[DEFAULT]**
     -
   * - ``lvm_conf_file`` = ``/etc/cinder/lvm.conf``
     - (String) LVM conf file to use for the LVM driver in Cinder; this setting is ignored if the specified file does not exist (You can also specify 'None' to not use a conf file even if one exists).
   * - ``lvm_max_over_subscription_ratio`` = ``1.0``
     - (Floating point) max_over_subscription_ratio setting for the LVM driver. If set, this takes precedence over the general max_over_subscription_ratio option. If None, the general option is used.
   * - ``lvm_mirrors`` = ``0``
     - (Integer) If >0, create LVs with multiple mirrors. Note that this requires lvm_mirrors + 2 PVs with available space
   * - ``lvm_suppress_fd_warnings`` = ``False``
     - (Boolean) Suppress leaked file descriptor warnings in LVM commands.
   * - ``lvm_type`` = ``auto``
     - (String) Type of LVM volumes to deploy; (default, thin, or auto). Auto defaults to thin if thin is supported.
   * - ``volume_group`` = ``cinder-volumes``
     - (String) Name for the VG that will contain exported volumes
