..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _cinder-hpelefthand:

.. list-table:: Description of HPE LeftHand/StoreVirtual driver configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[DEFAULT]**
     -
   * - ``hpelefthand_api_url`` = ``None``
     - (URI) HPE LeftHand WSAPI Server Url like https://<LeftHand ip>:8081/lhos
   * - ``hpelefthand_clustername`` = ``None``
     - (String) HPE LeftHand cluster name
   * - ``hpelefthand_debug`` = ``False``
     - (Boolean) Enable HTTP debugging to LeftHand
   * - ``hpelefthand_iscsi_chap_enabled`` = ``False``
     - (Boolean) Configure CHAP authentication for iSCSI connections (Default: Disabled)
   * - ``hpelefthand_password`` = ``None``
     - (String) HPE LeftHand Super user password
   * - ``hpelefthand_ssh_port`` = ``16022``
     - (Port number) Port number of SSH service.
   * - ``hpelefthand_username`` = ``None``
     - (String) HPE LeftHand Super user username
