..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _cinder-nas:

.. list-table:: Description of NAS configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[DEFAULT]**
     -
   * - ``nas_host`` =
     - (String) IP address or Hostname of NAS system.
   * - ``nas_login`` = ``admin``
     - (String) User name to connect to NAS system.
   * - ``nas_mount_options`` = ``None``
     - (String) Options used to mount the storage backend file system where Cinder volumes are stored.
   * - ``nas_password`` =
     - (String) Password to connect to NAS system.
   * - ``nas_private_key`` =
     - (String) Filename of private key to use for SSH authentication.
   * - ``nas_secure_file_operations`` = ``auto``
     - (String) Allow network-attached storage systems to operate in a secure environment where root level access is not permitted. If set to False, access is as the root user and insecure. If set to True, access is not as root. If set to auto, a check is done to determine if this is a new installation: True is used if so, otherwise False. Default is auto.
   * - ``nas_secure_file_permissions`` = ``auto``
     - (String) Set more secure file permissions on network-attached storage volume files to restrict broad other/world access. If set to False, volumes are created with open permissions. If set to True, volumes are created with permissions for the cinder user and group (660). If set to auto, a check is done to determine if this is a new installation: True is used if so, otherwise False. Default is auto.
   * - ``nas_share_path`` =
     - (String) Path to the share to use for storing Cinder volumes. For example: "/srv/export1" for an NFS server export available at 10.0.5.10:/srv/export1 .
   * - ``nas_ssh_port`` = ``22``
     - (Port number) SSH port to use to connect to NAS system.
   * - ``nas_volume_prov_type`` = ``thin``
     - (String) Provisioning type that will be used when creating volumes.
