..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _cinder-nec_m:

.. list-table:: Description of NEC Storage M series driver configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[DEFAULT]**
     -
   * - ``nec_actual_free_capacity`` = ``False``
     - (Boolean) Return actual free capacity.
   * - ``nec_backend_max_ld_count`` = ``1024``
     - (Integer) Maximum number of managing sessions.
   * - ``nec_backup_ldname_format`` = ``LX:%s``
     - (String) M-Series Storage LD name format for snapshots.
   * - ``nec_backup_pools`` =
     - (List) M-Series Storage backup pool number to be used.
   * - ``nec_diskarray_name`` =
     - (String) Diskarray name of M-Series Storage.
   * - ``nec_iscsi_portals_per_cont`` = ``1``
     - (Integer) Number of iSCSI portals.
   * - ``nec_ismcli_fip`` = ``None``
     - (IP) FIP address of M-Series Storage iSMCLI.
   * - ``nec_ismcli_password`` =
     - (String) Password for M-Series Storage iSMCLI.
   * - ``nec_ismcli_privkey`` =
     - (String) Filename of RSA private key for M-Series Storage iSMCLI.
   * - ``nec_ismcli_user`` =
     - (String) User name for M-Series Storage iSMCLI.
   * - ``nec_ismview_alloptimize`` = ``False``
     - (Boolean) Use legacy iSMCLI command with optimization.
   * - ``nec_ismview_dir`` = ``/tmp/nec/cinder``
     - (String) Output path of iSMview file.
   * - ``nec_ldname_format`` = ``LX:%s``
     - (String) M-Series Storage LD name format for volumes.
   * - ``nec_ldset`` =
     - (String) M-Series Storage LD Set name for Compute Node.
   * - ``nec_ldset_for_controller_node`` =
     - (String) M-Series Storage LD Set name for Controller Node.
   * - ``nec_pools`` =
     - (List) M-Series Storage pool numbers list to be used.
   * - ``nec_queryconfig_view`` = ``False``
     - (Boolean) Use legacy iSMCLI command.
   * - ``nec_ssh_pool_port_number`` = ``22``
     - (Integer) Port number of ssh pool.
   * - ``nec_unpairthread_timeout`` = ``3600``
     - (Integer) Timeout value of Unpairthread.
