..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _cinder-hpexp:

.. list-table:: Description of HPE XP volume driver configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[DEFAULT]**
     -
   * - ``hpexp_async_copy_check_interval`` = ``10``
     - (Integer) Interval to check copy asynchronously
   * - ``hpexp_compute_target_ports`` = ``None``
     - (List) Target port names of compute node for host group or iSCSI target
   * - ``hpexp_copy_check_interval`` = ``3``
     - (Integer) Interval to check copy
   * - ``hpexp_copy_speed`` = ``3``
     - (Integer) Copy speed of storage system
   * - ``hpexp_default_copy_method`` = ``FULL``
     - (String) Default copy method of storage system. There are two valid values: "FULL" specifies that a full copy; "THIN" specifies that a thin copy. Default value is "FULL"
   * - ``hpexp_group_request`` = ``False``
     - (Boolean) Request for creating host group or iSCSI target
   * - ``hpexp_horcm_add_conf`` = ``True``
     - (Boolean) Add to HORCM configuration
   * - ``hpexp_horcm_name_only_discovery`` = ``False``
     - (Boolean) Only discover a specific name of host group or iSCSI target
   * - ``hpexp_horcm_numbers`` = ``200, 201``
     - (List) Instance numbers for HORCM
   * - ``hpexp_horcm_resource_name`` = ``meta_resource``
     - (String) Resource group name of storage system for HORCM
   * - ``hpexp_horcm_user`` = ``None``
     - (String) Username of storage system for HORCM
   * - ``hpexp_ldev_range`` = ``None``
     - (String) Logical device range of storage system
   * - ``hpexp_pool`` = ``None``
     - (String) Pool of storage system
   * - ``hpexp_storage_cli`` = ``None``
     - (String) Type of storage command line interface
   * - ``hpexp_storage_id`` = ``None``
     - (String) ID of storage system
   * - ``hpexp_target_ports`` = ``None``
     - (List) Target port names for host group or iSCSI target
   * - ``hpexp_thin_pool`` = ``None``
     - (String) Thin pool of storage system
   * - ``hpexp_zoning_request`` = ``False``
     - (Boolean) Request for FC Zone creating host group
