..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _cinder-solidfire:

.. list-table:: Description of SolidFire driver configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[DEFAULT]**
     -
   * - ``sf_account_prefix`` = ``None``
     - (String) Create SolidFire accounts with this prefix. Any string can be used here, but the string "hostname" is special and will create a prefix using the cinder node hostname (previous default behavior). The default is NO prefix.
   * - ``sf_allow_template_caching`` = ``True``
     - (Boolean) Create an internal cache of copy of images when a bootable volume is created to eliminate fetch from glance and qemu-conversion on subsequent calls.
   * - ``sf_allow_tenant_qos`` = ``False``
     - (Boolean) Allow tenants to specify QOS on create
   * - ``sf_api_port`` = ``443``
     - (Port number) SolidFire API port. Useful if the device api is behind a proxy on a different port.
   * - ``sf_emulate_512`` = ``True``
     - (Boolean) Set 512 byte emulation on volume creation;
   * - ``sf_enable_vag`` = ``False``
     - (Boolean) Utilize volume access groups on a per-tenant basis.
   * - ``sf_enable_volume_mapping`` = ``True``
     - (Boolean) Create an internal mapping of volume IDs and account. Optimizes lookups and performance at the expense of memory, very large deployments may want to consider setting to False.
   * - ``sf_svip`` = ``None``
     - (String) Overrides default cluster SVIP with the one specified. This is required or deployments that have implemented the use of VLANs for iSCSI networks in their cloud.
   * - ``sf_template_account_name`` = ``openstack-vtemplate``
     - (String) Account name on the SolidFire Cluster to use as owner of template/cache volumes (created if does not exist).
   * - ``sf_volume_prefix`` = ``UUID-``
     - (String) Create SolidFire volumes with this prefix. Volume names are of the form <sf_volume_prefix><cinder-volume-id>. The default is to use a prefix of 'UUID-'.
