#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_db.sqlalchemy import utils
from oslo_log import log as logging
from sqlalchemy import MetaData

LOG = logging.getLogger(__name__)


def upgrade(migrate_engine):
    meta = MetaData()
    meta.bind = migrate_engine
    index_name = 'quota_usage_project_resource_idx'
    columns = ['project_id', 'resource']
    if utils.index_exists_on_columns(migrate_engine, 'quota_usages', columns):
        LOG.info(
            'Skipped adding %s because an equivalent index already exists.',
            index_name
        )
    else:
        utils.add_index(migrate_engine, 'quota_usages', index_name, columns)
