..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _cinder-netapp_eseries_iscsi:

.. list-table:: Description of NetApp E-Series driver configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[DEFAULT]**
     -
   * - ``netapp_controller_ips`` = ``None``
     - (String) This option is only utilized when the storage family is configured to eseries. This option is used to restrict provisioning to the specified controllers. Specify the value of this option to be a comma separated list of controller hostnames or IP addresses to be used for provisioning.
   * - ``netapp_enable_multiattach`` = ``False``
     - (Boolean) This option specifies whether the driver should allow operations that require multiple attachments to a volume. An example would be live migration of servers that have volumes attached. When enabled, this backend is limited to 256 total volumes in order to guarantee volumes can be accessed by more than one host.
   * - ``netapp_host_type`` = ``None``
     - (String) This option defines the type of operating system for all initiators that can access a LUN. This information is used when mapping LUNs to individual hosts or groups of hosts.
   * - ``netapp_login`` = ``None``
     - (String) Administrative user account name used to access the storage system or proxy server.
   * - ``netapp_password`` = ``None``
     - (String) Password for the administrative user account specified in the netapp_login option.
   * - ``netapp_pool_name_search_pattern`` = ``(.+)``
     - (String) This option is used to restrict provisioning to the specified pools. Specify the value of this option to be a regular expression which will be applied to the names of objects from the storage backend which represent pools in Cinder. This option is only utilized when the storage protocol is configured to use iSCSI or FC.
   * - ``netapp_replication_aggregate_map`` = ``None``
     - (Unknown) Multi opt of dictionaries to represent the aggregate mapping between source and destination back ends when using whole back end replication. For every source aggregate associated with a cinder pool (NetApp FlexVol), you would need to specify the destination aggregate on the replication target device. A replication target device is configured with the configuration option replication_device. Specify this option as many times as you have replication devices. Each entry takes the standard dict config form: netapp_replication_aggregate_map = backend_id:<name_of_replication_device_section>,src_aggr_name1:dest_aggr_name1,src_aggr_name2:dest_aggr_name2,...
   * - ``netapp_sa_password`` = ``None``
     - (String) Password for the NetApp E-Series storage array.
   * - ``netapp_server_hostname`` = ``None``
     - (String) The hostname (or IP address) for the storage system or proxy server.
   * - ``netapp_server_port`` = ``None``
     - (Integer) The TCP port to use for communication with the storage system or proxy server. If not specified, Data ONTAP drivers will use 80 for HTTP and 443 for HTTPS; E-Series will use 8080 for HTTP and 8443 for HTTPS.
   * - ``netapp_snapmirror_quiesce_timeout`` = ``3600``
     - (Integer) The maximum time in seconds to wait for existing SnapMirror transfers to complete before aborting during a failover.
   * - ``netapp_storage_family`` = ``ontap_cluster``
     - (String) The storage family type used on the storage system; valid values are ontap_cluster for using clustered Data ONTAP, or eseries for using E-Series.
   * - ``netapp_transport_type`` = ``http``
     - (String) The transport protocol used when communicating with the storage system or proxy server.
   * - ``netapp_webservice_path`` = ``/devmgr/v2``
     - (String) This option is used to specify the path to the E-Series proxy application on a proxy server. The value is combined with the value of the netapp_transport_type, netapp_server_hostname, and netapp_server_port options to create the URL used by the driver to connect to the proxy application.
