..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _cinder-disco:

.. list-table:: Description of Disco volume driver configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[DEFAULT]**
     -
   * - ``choice_client`` = ``None``
     - (String) Use soap client or rest client for communicating with DISCO. Possible values are "soap" or "rest".
   * - ``clone_check_timeout`` = ``3600``
     - (Integer) How long we check whether a clone is finished before we give up
   * - ``clone_volume_timeout`` = ``680``
     - (Integer) Create clone volume timeout.
   * - ``disco_client`` = ``127.0.0.1``
     - (IP) The IP of DMS client socket server
   * - ``disco_client_port`` = ``9898``
     - (Port number) The port to connect DMS client socket server
   * - ``disco_src_api_port`` = ``8080``
     - (Port number) The port of DISCO source API
   * - ``disco_wsdl_path`` = ``/etc/cinder/DISCOService.wsdl``
     - (String) DEPRECATED: Path to the wsdl file to communicate with DISCO request manager
   * - ``rest_ip`` = ``None``
     - (IP) The IP address of the REST server
   * - ``restore_check_timeout`` = ``3600``
     - (Integer) How long we check whether a restore is finished before we give up
   * - ``retry_interval`` = ``1``
     - (Integer) How long we wait before retrying to get an item detail
