..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _cinder-dellsc:

.. list-table:: Description of Dell Storage Center volume driver configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[DEFAULT]**
     -
   * - ``dell_sc_api_port`` = ``3033``
     - (Port number) Dell API port
   * - ``dell_sc_server_folder`` = ``openstack``
     - (String) Name of the server folder to use on the Storage Center
   * - ``dell_sc_ssn`` = ``64702``
     - (Integer) Storage Center System Serial Number
   * - ``dell_sc_verify_cert`` = ``False``
     - (Boolean) Enable HTTPS SC certificate verification
   * - ``dell_sc_volume_folder`` = ``openstack``
     - (String) Name of the volume folder to use on the Storage Center
   * - ``dell_server_os`` = ``Red Hat Linux 6.x``
     - (String) Server OS type to use when creating a new server on the Storage Center.
   * - ``excluded_domain_ip`` = ``None``
     - (Unknown) Domain IP to be excluded from iSCSI returns.
   * - ``secondary_san_ip`` =
     - (String) IP address of secondary DSM controller
   * - ``secondary_san_login`` = ``Admin``
     - (String) Secondary DSM user name
   * - ``secondary_san_password`` =
     - (String) Secondary DSM user password name
   * - ``secondary_sc_api_port`` = ``3033``
     - (Port number) Secondary Dell API port
