..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _cinder-compute:

.. list-table:: Description of Compute configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[DEFAULT]**
     -
   * - ``nova_api_insecure`` = ``False``
     - (Boolean) Allow to perform insecure SSL requests to nova
   * - ``nova_ca_certificates_file`` = ``None``
     - (String) Location of ca certificates file to use for nova client requests.
   * - ``nova_catalog_admin_info`` = ``compute:Compute Service:adminURL``
     - (String) Same as nova_catalog_info, but for admin endpoint.
   * - ``nova_catalog_info`` = ``compute:Compute Service:publicURL``
     - (String) Match this value when searching for nova in the service catalog. Format is: separated values of the form: <service_type>:<service_name>:<endpoint_type>
   * - ``nova_endpoint_admin_template`` = ``None``
     - (String) Same as nova_endpoint_template, but for admin endpoint.
   * - ``nova_endpoint_template`` = ``None``
     - (String) Override service catalog lookup with template for nova endpoint e.g. http://localhost:8774/v2/%(project_id)s
   * - ``os_region_name`` = ``None``
     - (String) Region name of this node
