..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _cinder-backups:

.. list-table:: Description of backups configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[DEFAULT]**
     -
   * - ``backup_api_class`` = ``cinder.backup.api.API``
     - (String) The full class name of the volume backup API class
   * - ``backup_compression_algorithm`` = ``zlib``
     - (String) Compression algorithm (None to disable)
   * - ``backup_driver`` = ``cinder.backup.drivers.swift``
     - (String) Driver to use for backups.
   * - ``backup_manager`` = ``cinder.backup.manager.BackupManager``
     - (String) Full class name for the Manager for volume backup
   * - ``backup_metadata_version`` = ``2``
     - (Integer) Backup metadata version to be used when backing up volume metadata. If this number is bumped, make sure the service doing the restore supports the new version.
   * - ``backup_name_template`` = ``backup-%s``
     - (String) Template string to be used to generate backup names
   * - ``backup_object_number_per_notification`` = ``10``
     - (Integer) The number of chunks or objects, for which one Ceilometer notification will be sent
   * - ``backup_service_inithost_offload`` = ``True``
     - (Boolean) Offload pending backup delete during backup service startup. If false, the backup service will remain down until all pending backups are deleted.
   * - ``backup_timer_interval`` = ``120``
     - (Integer) Interval, in seconds, between two progress notifications reporting the backup status
   * - ``backup_use_same_host`` = ``False``
     - (Boolean) Backup services use same backend.
   * - ``backup_use_temp_snapshot`` = ``False``
     - (Boolean) If this is set to True, the backup_use_temp_snapshot path will be used during the backup. Otherwise, it will use backup_use_temp_volume path.
   * - ``snapshot_check_timeout`` = ``3600``
     - (Integer) How long we check whether a snapshot is finished before we give up
   * - ``snapshot_name_template`` = ``snapshot-%s``
     - (String) Template string to be used to generate snapshot names
   * - ``snapshot_same_host`` = ``True``
     - (Boolean) Create volume from snapshot at the host where snapshot resides
