..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _cinder-emc_sio:

.. list-table:: Description of EMC SIO volume driver configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description

   * - **[DEFAULT]**
     -

   * - ``sio_max_over_subscription_ratio`` = ``10.0``

     - (Floating point) max_over_subscription_ratio setting for the ScaleIO driver. This replaces the general max_over_subscription_ratio which has no effect in this driver.Maximum value allowed for ScaleIO is 10.0.

   * - ``sio_protection_domain_id`` = ``None``

     - (String) Protection Domain ID.

       - **Deprecated**

         Replaced by sio_storage_pools option

   * - ``sio_protection_domain_name`` = ``None``

     - (String) Protection Domain name.

       - **Deprecated**

         Replaced by sio_storage_pools option

   * - ``sio_rest_server_port`` = ``443``

     - (String) REST server port.

   * - ``sio_round_volume_capacity`` = ``True``

     - (Boolean) Round up volume capacity.

   * - ``sio_server_certificate_path`` = ``None``

     - (String) Server certificate path.

   * - ``sio_storage_pool_id`` = ``None``

     - (String) Storage Pool ID.

       - **Deprecated**

         Replaced by sio_storage_pools option

   * - ``sio_storage_pool_name`` = ``None``

     - (String) Storage Pool name.

       - **Deprecated**

         Replaced by sio_storage_pools option

   * - ``sio_storage_pools`` = ``None``

     - (String) Storage Pools.

   * - ``sio_unmap_volume_before_deletion`` = ``False``

     - (Boolean) Unmap volume before deletion.

   * - ``sio_verify_server_certificate`` = ``False``

     - (Boolean) Verify server certificate.
