..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _cinder-emc:

.. list-table:: Description of EMC configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[DEFAULT]**
     -
   * - ``check_max_pool_luns_threshold`` = ``False``
     - (Boolean) Report free_capacity_gb as 0 when the limit to maximum number of pool LUNs is reached. By default, the value is False.
   * - ``cinder_emc_config_file`` = ``/etc/cinder/cinder_emc_config.xml``
     - (String) Use this file for cinder emc plugin config data
   * - ``destroy_empty_storage_group`` = ``False``
     - (Boolean) To destroy storage group when the last LUN is removed from it. By default, the value is False.
   * - ``force_delete_lun_in_storagegroup`` = ``False``
     - (Boolean) Delete a LUN even if it is in Storage Groups. By default, the value is False.
   * - ``initiator_auto_deregistration`` = ``False``
     - (Boolean) Automatically deregister initiators after the related storage group is destroyed. By default, the value is False.
   * - ``initiator_auto_registration`` = ``False``
     - (Boolean) Automatically register initiators. By default, the value is False.
   * - ``io_port_list`` = ``None``
     - (List) Comma separated iSCSI or FC ports to be used in Nova or Cinder.
   * - ``iscsi_initiators`` = ``None``
     - (String) Mapping between hostname and its iSCSI initiator IP addresses.
   * - ``max_luns_per_storage_group`` = ``255``
     - (Integer) Default max number of LUNs in a storage group. By default, the value is 255.
   * - ``multi_pool_support`` = ``False``
     - (String) Use this value to specify multi-pool support for VMAX3
   * - ``naviseccli_path`` = ``None``
     - (String) Naviseccli Path.
   * - ``storage_vnx_authentication_type`` = ``global``
     - (String) VNX authentication scope type. By default, the value is global.
   * - ``storage_vnx_pool_names`` = ``None``
     - (List) Comma-separated list of storage pool names to be used.
   * - ``storage_vnx_security_file_dir`` = ``None``
     - (String) Directory path that contains the VNX security file. Make sure the security file is generated first.
