..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _cinder-zoning:

.. list-table:: Description of zoning configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[DEFAULT]**
     -
   * - ``zoning_mode`` = ``None``
     - (String) FC Zoning mode configured
   * - **[fc-zone-manager]**
     -
   * - ``enable_unsupported_driver`` = ``False``
     - (Boolean) Set this to True when you want to allow an unsupported zone manager driver to start. Drivers that haven't maintained a working CI system and testing are marked as unsupported until CI is working again. This also marks a driver as deprecated and may be removed in the next release.
   * - ``fc_fabric_names`` = ``None``
     - (String) Comma separated list of Fibre Channel fabric names. This list of names is used to retrieve other SAN credentials for connecting to each SAN fabric
   * - ``fc_san_lookup_service`` = ``cinder.zonemanager.drivers.brocade.brcd_fc_san_lookup_service.BrcdFCSanLookupService``
     - (String) FC SAN Lookup Service
   * - ``zone_driver`` = ``cinder.zonemanager.drivers.brocade.brcd_fc_zone_driver.BrcdFCZoneDriver``
     - (String) FC Zone Driver responsible for zone management
   * - ``zoning_policy`` = ``initiator-target``
     - (String) Zoning policy configured by user; valid values include "initiator-target" or "initiator"
