..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _cinder-flashsystem:

.. list-table:: Description of IBM FlashSystem volume driver configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[DEFAULT]**
     -
   * - ``flashsystem_connection_protocol`` = ``FC``
     - (String) Connection protocol should be FC. (Default is FC.)
   * - ``flashsystem_iscsi_portid`` = ``0``
     - (Integer) Default iSCSI Port ID of FlashSystem. (Default port is 0.)
   * - ``flashsystem_multihostmap_enabled`` = ``True``
     - (Boolean) Allows vdisk to multi host mapping. (Default is True)
   * - ``flashsystem_multipath_enabled`` = ``False``
     - (Boolean) DEPRECATED: This option no longer has any affect. It is deprecated and will be removed in the next release.
