..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _cinder-zoning_fabric_cisco:

.. list-table:: Description of cisco zoning fabrics configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[CISCO_FABRIC_EXAMPLE]**
     -
   * - ``cisco_fc_fabric_address`` =
     - (String) Management IP of fabric
   * - ``cisco_fc_fabric_password`` =
     - (String) Password for user
   * - ``cisco_fc_fabric_port`` = ``22``
     - (Port number) Connecting port
   * - ``cisco_fc_fabric_user`` =
     - (String) Fabric user ID
   * - ``cisco_zone_activate`` = ``True``
     - (Boolean) overridden zoning activation state
   * - ``cisco_zone_name_prefix`` = ``None``
     - (String) overridden zone name prefix
   * - ``cisco_zoning_policy`` = ``initiator-target``
     - (String) overridden zoning policy
   * - ``cisco_zoning_vsan`` = ``None``
     - (String) VSAN of the Fabric
