..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _cinder-storage_ceph:

.. list-table:: Description of Ceph storage configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[DEFAULT]**
     -
   * - ``rados_connect_timeout`` = ``-1``
     - (Integer) Timeout value (in seconds) used when connecting to ceph cluster. If value < 0, no timeout is set and default librados value is used.
   * - ``rados_connection_interval`` = ``5``
     - (Integer) Interval value (in seconds) between connection retries to ceph cluster.
   * - ``rados_connection_retries`` = ``3``
     - (Integer) Number of retries if connection to ceph cluster failed.
   * - ``rbd_ceph_conf`` =
     - (String) Path to the ceph configuration file
   * - ``rbd_cluster_name`` = ``ceph``
     - (String) The name of ceph cluster
   * - ``rbd_flatten_volume_from_snapshot`` = ``False``
     - (Boolean) Flatten volumes created from snapshots to remove dependency from volume to snapshot
   * - ``rbd_max_clone_depth`` = ``5``
     - (Integer) Maximum number of nested volume clones that are taken before a flatten occurs. Set to 0 to disable cloning.
   * - ``rbd_pool`` = ``rbd``
     - (String) The RADOS pool where rbd volumes are stored
   * - ``rbd_secret_uuid`` = ``None``
     - (String) The libvirt uuid of the secret for the rbd_user volumes
   * - ``rbd_store_chunk_size`` = ``4``
     - (Integer) Volumes will be chunked into objects of this size (in megabytes).
   * - ``rbd_user`` = ``None``
     - (String) The RADOS client name for accessing rbd volumes - only set when using cephx authentication
   * - ``replication_connect_timeout`` = ``5``
     - (Integer) Timeout value (in seconds) used when connecting to ceph cluster to do a demotion/promotion of volumes. If value < 0, no timeout is set and default librados value is used.
